/**
 * @copyright   2017-2018, Miles Johnson
 * @license     https://opensource.org/licenses/MIT
 */
import { Options as ExecaOptions, SyncOptions as ExecaSyncOptions, ExecaChildProcess, ExecaReturns } from 'execa';
import { Struct } from 'optimal';
import Context from './Context';
import Task, { TaskAction, TaskInterface } from './Task';
import { ToolInterface } from './Tool';
import { Debugger } from './types';
import { AggregatedResponse } from './Executor';
import { PoolExecutorOptions } from './executors/Pool';
export interface CommandOptions extends Struct {
    sync?: boolean;
    task?: TaskInterface;
    wrap?: (process: ExecaChildProcess) => void;
}
export interface RoutineInterface extends TaskInterface {
    key: string;
    routines: RoutineInterface[];
    tool: ToolInterface;
    run<T>(context: Context, initialValue?: T | null, wasParallel?: boolean): Promise<any>;
}
export default class Routine<To extends Struct, Tx extends Context> extends Task<To, Tx> implements RoutineInterface {
    exit: boolean;
    debug: Debugger;
    key: string;
    routines: RoutineInterface[];
    tool: ToolInterface;
    constructor(key: string, title: string, options?: Partial<To>);
    /**
     * Called once the routine has been configured and is ready to execute.
     */
    bootstrap(): void;
    /**
     * Configure the routine after it has been instantiated.
     */
    configure(parent: Routine<Struct, Tx>): this;
    /**
     * Execute the current routine and return a new value.
     * This method *must* be overridden in a subclass.
     */
    execute<T>(context: Tx, value?: T): Promise<any>;
    /**
     * Execute a command with the given arguments and pass the results through a promise.
     */
    executeCommand(command: string, args: string[], options?: (ExecaOptions | ExecaSyncOptions) & CommandOptions): Promise<ExecaReturns>;
    /**
     * Execute routines in parallel.
     */
    parallelizeRoutines<T>(value?: T, routines?: RoutineInterface[]): Promise<any[]>;
    /**
     * Execute tasks in parallel.
     */
    parallelizeTasks<T>(value?: T, tasks?: TaskInterface[]): Promise<any[]>;
    /**
     * Add a new routine within this routine.
     */
    pipe(routine: RoutineInterface): this;
    /**
     * Execute routines in a pool.
     */
    poolRoutines<T>(value?: T, options?: Partial<PoolExecutorOptions>, routines?: RoutineInterface[]): Promise<AggregatedResponse>;
    /**
     * Execute tasks in a pool.
     */
    poolTasks<T>(value?: T, options?: Partial<PoolExecutorOptions>, tasks?: TaskInterface[]): Promise<AggregatedResponse>;
    /**
     * Trigger processes before and after execution.
     */
    run<T>(context: Tx, value?: T, wasParallel?: boolean): Promise<any>;
    /**
     * Execute routines in sequential (serial) order.
     */
    serializeRoutines<T>(value?: T, routines?: RoutineInterface[]): Promise<any>;
    /**
     * Execute tasks in sequential (serial) order.
     */
    serializeTasks<T>(value?: T, tasks?: TaskInterface[]): Promise<any>;
    /**
     * Execute routines in sync.
     */
    synchronizeRoutines<T>(value?: T, routines?: RoutineInterface[]): Promise<AggregatedResponse>;
    /**
     * Execute tasks in sync.
     */
    synchronizeTasks<T>(value?: T, tasks?: TaskInterface[]): Promise<AggregatedResponse>;
    /**
     * Define an individual task.
     */
    task<Tp extends Struct>(title: string, action: TaskAction<Tx>, options?: Tp): TaskInterface;
}
