/**
 * @copyright   2017-2018, Miles Johnson
 * @license     https://opensource.org/licenses/MIT
 */
import { Blueprint, Struct } from 'optimal';
import { ConsoleInterface, ConsoleOptions } from './Console';
import Emitter, { EmitterInterface } from './Emitter';
import { PluginInterface } from './Plugin';
import { ReporterInterface } from './Reporter';
import { Debugger, ToolConfig, PackageConfig } from './types';
export interface ToolOptions extends Struct {
    appName: string;
    configBlueprint: Blueprint;
    configFolder: string;
    console: Partial<ConsoleOptions>;
    pluginAlias: string;
    root: string;
    scoped: boolean;
    workspaceRoot: string;
}
export interface ToolInterface extends EmitterInterface {
    argv: string[];
    config: ToolConfig;
    console: ConsoleInterface;
    debug: Debugger;
    options: ToolOptions;
    package: PackageConfig;
    plugins: PluginInterface[];
    reporters: ReporterInterface[];
    createDebugger(...namespaces: string[]): Debugger;
    initialize(): this;
    getPlugin(name: string): PluginInterface;
    getReporter(name: string): ReporterInterface;
    getThemeList(): string[];
    log(message: string, ...args: any[]): this;
    logError(message: string, ...args: any[]): this;
}
export default class Tool<Tp extends PluginInterface, Tr extends ReporterInterface> extends Emitter implements ToolInterface {
    argv: string[];
    config: ToolConfig;
    console: ConsoleInterface;
    debug: Debugger;
    initialized: boolean;
    options: ToolOptions;
    package: PackageConfig;
    plugins: Tp[];
    reporters: Tr[];
    constructor(options: Partial<ToolOptions>, argv?: string[]);
    /**
     * Add a reporter and bootstrap with the console instance.
     */
    addReporter(reporter: Tr): this;
    /**
     * Create a debugger with a namespace.
     */
    createDebugger(...namespaces: string[]): Debugger;
    /**
     * Force exit the application.
     */
    exit(message: string | Error | null, code?: number): this;
    /**
     * Get a plugin by name.
     */
    getPlugin(name: string): Tp;
    /**
     * Get a reporter by name.
     */
    getReporter(name: string): Tr;
    /**
     * Return a list of all theme names.
     */
    getThemeList(): string[];
    /**
     * Initialize the tool by loading config and plugins.
     */
    initialize(): this;
    /**
     * Load the package.json and local configuration files.
     *
     * Must be called first in the lifecycle.
     */
    loadConfig(): this;
    /**
     * Register plugins from the loaded configuration.
     *
     * Must be called after config has been loaded.
     */
    loadPlugins(): this;
    /**
     * Register reporters from the loaded configuration.
     *
     * Must be called after config has been loaded.
     */
    loadReporters(): this;
    /**
     * Log a message to the console to display on success.
     */
    log(message: string, ...args: any[]): this;
    /**
     * Log an error to the console to display on failure.
     */
    logError(message: string, ...args: any[]): this;
}
